/*
 * snul.vala is a part of Snul
 *
 * Snul: is a simple library that implements
 *       a tcp socket client in vala
 *
 * Copyright (C) 2008  Andrea Del Signore
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 *
 * Author:
 *     Andrea Del Signore <sejerpz@tin.it>
 */

#include <src/snul.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>

struct _SnulTcpClientPrivate {
	char* _address;
	char* _first_address;
	char* _port;
	gushort _port_number;
	gint _socket_fd;
	GIOChannel* _io_channel;
	guint _watch_id;
	guint _resolver_watch_id;
	gboolean _resolving_address;
	gboolean _resolv_timeout;
	struct hostent* hp;
};
#define SNUL_TCP_CLIENT_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), SNUL_TYPE_TCP_CLIENT, SnulTcpClientPrivate))
enum  {
	SNUL_TCP_CLIENT_DUMMY_PROPERTY,
	SNUL_TCP_CLIENT_ADDRESS,
	SNUL_TCP_CLIENT_FIRST_ADDRESS,
	SNUL_TCP_CLIENT_PORT
};
static gint snul_tcp_client_BUFFER_LENGTH = 1024;
static gchar* snul_tcp_client__io_channel_buffer;
static gint snul_tcp_client__io_channel_buffer_length1;
static const char* SNUL_TCP_CLIENT_error_domain_quark = "tcp_socket_error_domain-quark";
static gboolean snul_tcp_client_resolv_watcher (SnulTcpClient* self);
static gboolean _snul_tcp_client_resolv_watcher_gsource_func (gpointer self);
static void* snul_tcp_client_resolv_thread (SnulTcpClient* self);
static void* _snul_tcp_client_resolv_thread_gthread_func (gpointer self);
static void snul_tcp_client_resolv (SnulTcpClient* self, const char* address);
static gboolean _snul_tcp_client_socket_callback_utils_io_func (GIOChannel* source, GIOCondition conditions, gpointer data);
static void snul_tcp_client_connect_second_phase (SnulTcpClient* self);
static void snul_tcp_client_on_connected_real (SnulTcpClient* self, GError** error);
static gboolean snul_tcp_client_socket_callback (GIOChannel* channel, GIOCondition condition, gpointer data);
static gpointer snul_tcp_client_parent_class = NULL;
static void snul_tcp_client_dispose (GObject * obj);

static void g_cclosure_user_marshal_VOID__STRING_ULONG (GClosure * closure, GValue * return_value, guint n_param_values, const GValue * param_values, gpointer invocation_hint, gpointer marshal_data);

GQuark snul_tcp_socket_error_quark (void) {
	return g_quark_from_static_string ("snul_tcp_socket_error-quark");
}


static gboolean snul_tcp_client_resolv_watcher (SnulTcpClient* self) {
	g_return_val_if_fail (SNUL_IS_TCP_CLIENT (self), FALSE);
	if (self->priv->_resolving_address) {
		return TRUE;
	}
	snul_tcp_client_connect_second_phase (self);
	self->priv->_resolver_watch_id = ((guint) 0);
	return FALSE;
}


static gboolean _snul_tcp_client_resolv_watcher_gsource_func (gpointer self) {
	return snul_tcp_client_resolv_watcher (self);
}


static void* snul_tcp_client_resolv_thread (SnulTcpClient* self) {
	g_return_val_if_fail (SNUL_IS_TCP_CLIENT (self), NULL);
	self->priv->_resolving_address = TRUE;
	self->priv->_resolver_watch_id = g_timeout_add (((guint) 250), _snul_tcp_client_resolv_watcher_gsource_func, self);
	self->priv->hp = gethostbyname (self->priv->_address);
	if (self->priv->hp == NULL) {
		char* _tmp0;
		_tmp0 = NULL;
		self->priv->_first_address = (_tmp0 = NULL, (self->priv->_first_address = (g_free (self->priv->_first_address), NULL)), _tmp0);
	} else {
		char* _tmp1;
		_tmp1 = NULL;
		self->priv->_first_address = (_tmp1 = g_strdup_printf ("%s", self->priv->hp->h_addr_list[0]), (self->priv->_first_address = (g_free (self->priv->_first_address), NULL)), _tmp1);
	}
	self->priv->_resolving_address = FALSE;
	return NULL;
}


static void* _snul_tcp_client_resolv_thread_gthread_func (gpointer self) {
	return snul_tcp_client_resolv_thread (self);
}


static void snul_tcp_client_resolv (SnulTcpClient* self, const char* address) {
	GError * inner_error;
	g_return_if_fail (SNUL_IS_TCP_CLIENT (self));
	inner_error = NULL;
	{
		g_thread_create (_snul_tcp_client_resolv_thread_gthread_func, self, FALSE, &inner_error);
		if (inner_error != NULL) {
			if (inner_error->domain == G_THREAD_ERROR) {
				goto __catch1_g_thread_error;
			}
			g_critical ("file %s: line %d: uncaught error: %s", __FILE__, __LINE__, inner_error->message);
			g_clear_error (&inner_error);
		}
	}
	goto __finally1;
	__catch1_g_thread_error:
	{
		GError * err;
		err = inner_error;
		inner_error = NULL;
		{
			g_warning ("snul.vala:105: resolv error %d: %s", err->code, err->message);
		}
	}
	__finally1:
	;
}


void snul_tcp_client_connect (SnulTcpClient* self, const char* address, const char* port, GError** error) {
	GError * inner_error;
	g_return_if_fail (SNUL_IS_TCP_CLIENT (self));
	inner_error = NULL;
	{
		char* _tmp1;
		const char* _tmp0;
		char* _tmp3;
		const char* _tmp2;
		_tmp1 = NULL;
		_tmp0 = NULL;
		self->priv->_address = (_tmp1 = (_tmp0 = address, (_tmp0 == NULL ? NULL : g_strdup (_tmp0))), (self->priv->_address = (g_free (self->priv->_address), NULL)), _tmp1);
		_tmp3 = NULL;
		_tmp2 = NULL;
		self->priv->_port = (_tmp3 = (_tmp2 = port, (_tmp2 == NULL ? NULL : g_strdup (_tmp2))), (self->priv->_port = (g_free (self->priv->_port), NULL)), _tmp3);
		if (self->priv->_socket_fd > 0) {
			inner_error = g_error_new (SNUL_TCP_SOCKET_ERROR, SNUL_TCP_SOCKET_ERROR_SOCKET_ALREADY_OPEN, "Socket already opened");
			if (inner_error != NULL) {
				goto __catch2_g_error;
			}
		}
		if (g_unichar_isdigit (g_utf8_get_char (g_utf8_offset_to_pointer (self->priv->_port, 0)))) {
			self->priv->_port_number = ((gushort) atoi (self->priv->_port));
		} else {
			struct servent* sp;
			sp = getservbyname (self->priv->_port, "tcp");
			if (sp == NULL) {
				inner_error = g_error_new (SNUL_TCP_SOCKET_ERROR, SNUL_TCP_SOCKET_ERROR_LOOKUP_SERVICE_FAILED, "Lookup service by name failed");
				if (inner_error != NULL) {
					goto __catch2_g_error;
				}
			}
			self->priv->_port_number = ((gushort) sp->s_port);
		}
		snul_tcp_client_resolv (self, address);
	}
	goto __finally2;
	__catch2_g_error:
	{
		GError * err;
		err = inner_error;
		inner_error = NULL;
		{
			snul_tcp_client_on_error (self, ((GError*) err));
			inner_error = err;
			if (inner_error != NULL) {
				goto __catch2_g_error;
			}
		}
	}
	__finally2:
	;
}


static gboolean _snul_tcp_client_socket_callback_utils_io_func (GIOChannel* source, GIOCondition conditions, gpointer data) {
	return snul_tcp_client_socket_callback (source, conditions, data);
}


static void snul_tcp_client_connect_second_phase (SnulTcpClient* self) {
	GError * inner_error;
	g_return_if_fail (SNUL_IS_TCP_CLIENT (self));
	inner_error = NULL;
	{
		struct sockaddr_in* sin;
		struct in_addr* add;
		GIOChannel* _tmp0;
		GIOCondition conditions;
		if (self->priv->hp == NULL) {
			inner_error = g_error_new (SNUL_TCP_SOCKET_ERROR, SNUL_TCP_SOCKET_ERROR_LOOKUP_HOSTNAME_FAILED, "Resolv failed: hostname not found");
			if (inner_error != NULL) {
				goto __catch3_g_error;
			}
		}
		self->priv->_socket_fd = socket (AF_INET, ((gint) SOCK_STREAM), 0);
		if (self->priv->_socket_fd < 0) {
			inner_error = g_error_new (SNUL_TCP_SOCKET_ERROR, SNUL_TCP_SOCKET_ERROR_SOCKET_CREATION_ERROR, "Socket creation failed");
			if (inner_error != NULL) {
				goto __catch3_g_error;
			}
		}
		sin = g_new0 (struct sockaddr_in, 1);
		add = self->priv->hp->h_addr_list[0];
		sin->sin_family = ((gushort) AF_INET);
		sin->sin_addr.s_addr = add->s_addr;
		sin->sin_port = ((guint16) htons (self->priv->_port_number));
		if (connect (self->priv->_socket_fd, ((struct sockaddr*) sin), ((guint) sizeof (struct sockaddr))) < 0) {
			inner_error = g_error_new (SNUL_TCP_SOCKET_ERROR, SNUL_TCP_SOCKET_ERROR_SOCKET_CONNECT_ERROR, "Connect error");
			if (inner_error != NULL) {
				goto __catch3_g_error;
			}
		}
		if (self->priv->_socket_fd <= 0) {
			inner_error = g_error_new (SNUL_TCP_SOCKET_ERROR, SNUL_TCP_SOCKET_ERROR_SOCKET_CLOSED, "Closed or bad socket");
			if (inner_error != NULL) {
				goto __catch3_g_error;
			}
		}
		_tmp0 = NULL;
		self->priv->_io_channel = (_tmp0 = g_io_channel_unix_new (self->priv->_socket_fd), (self->priv->_io_channel == NULL ? NULL : (self->priv->_io_channel = (g_io_channel_unref (self->priv->_io_channel), NULL))), _tmp0);
		{
			g_io_channel_set_encoding (self->priv->_io_channel, NULL, &inner_error);
			if (inner_error != NULL) {
				if (inner_error->domain == G_CONVERT_ERROR) {
					goto __catch4_g_convert_error;
				}
				g_critical ("file %s: line %d: uncaught error: %s", __FILE__, __LINE__, inner_error->message);
				g_clear_error (&inner_error);
			}
			g_io_channel_set_flags (self->priv->_io_channel, G_IO_FLAG_APPEND | G_IO_FLAG_NONBLOCK, &inner_error);
			if (inner_error != NULL) {
				if (inner_error->domain == G_CONVERT_ERROR) {
					goto __catch4_g_convert_error;
				}
				g_critical ("file %s: line %d: uncaught error: %s", __FILE__, __LINE__, inner_error->message);
				g_clear_error (&inner_error);
			}
		}
		goto __finally4;
		__catch4_g_convert_error:
		{
			GError * err;
			err = inner_error;
			inner_error = NULL;
			{
				inner_error = err;
				if (inner_error != NULL) {
					if (inner_error->domain == G_CONVERT_ERROR) {
						goto __catch4_g_convert_error;
					}
					g_critical ("file %s: line %d: uncaught error: %s", __FILE__, __LINE__, inner_error->message);
					g_clear_error (&inner_error);
				}
			}
		}
		__finally4:
		;
		conditions = G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL;
		self->priv->_watch_id = g_io_add_watch (self->priv->_io_channel, conditions, _snul_tcp_client_socket_callback_utils_io_func, ((gpointer) self));
		snul_tcp_client_on_connected (self, self->priv->_socket_fd);
		(sin == NULL ? NULL : (sin = (free (sin), NULL)));
	}
	goto __finally3;
	__catch3_g_error:
	{
		GError * err;
		err = inner_error;
		inner_error = NULL;
		{
			snul_tcp_client_on_error (self, ((GError*) err));
		}
	}
	__finally3:
	;
}


gsize snul_tcp_client_send (SnulTcpClient* self, const char* buffer) {
	GError * inner_error;
	gssize len;
	gsize bytes_written;
	g_return_val_if_fail (SNUL_IS_TCP_CLIENT (self), 0UL);
	inner_error = NULL;
	if (buffer == NULL) {
		return ((gsize) 0);
	}
	len = ((gssize) strlen (buffer));
	bytes_written = 0UL;
	{
		g_io_channel_write_chars (self->priv->_io_channel, ((gchar*) buffer), ((gssize) len), &bytes_written, &inner_error);
		if (inner_error != NULL) {
			goto __catch5_g_error;
		}
		g_io_channel_flush (self->priv->_io_channel, &inner_error);
		if (inner_error != NULL) {
			goto __catch5_g_error;
		}
	}
	goto __finally5;
	__catch5_g_error:
	{
		GError * err;
		err = inner_error;
		inner_error = NULL;
		{
			g_warning ("snul.vala:190: error writing: %s", err->message);
		}
	}
	__finally5:
	;
	return bytes_written;
}


void snul_tcp_client_disconnect (SnulTcpClient* self) {
	GError * inner_error;
	g_return_if_fail (SNUL_IS_TCP_CLIENT (self));
	inner_error = NULL;
	if (self->priv->_watch_id != 0) {
		g_source_remove (self->priv->_watch_id);
		self->priv->_watch_id = ((guint) 0);
	}
	if (self->priv->_io_channel != NULL) {
		GIOChannel* _tmp0;
		{
			g_io_channel_shutdown (self->priv->_io_channel, FALSE, &inner_error);
			if (inner_error != NULL) {
				goto __catch6_g_error;
			}
		}
		goto __finally6;
		__catch6_g_error:
		{
			GError * err;
			err = inner_error;
			inner_error = NULL;
			{
				g_warning ("snul.vala:206: error shutting down channel: %s", err->message);
			}
		}
		__finally6:
		;
		_tmp0 = NULL;
		self->priv->_io_channel = (_tmp0 = NULL, (self->priv->_io_channel == NULL ? NULL : (self->priv->_io_channel = (g_io_channel_unref (self->priv->_io_channel), NULL))), _tmp0);
	}
	if (self->priv->_socket_fd != 0) {
		close (self->priv->_socket_fd);
	}
	snul_tcp_client_on_disconnected (self);
}


void snul_tcp_client_on_connected (SnulTcpClient* self, gint socket_fd) {
	g_return_if_fail (SNUL_IS_TCP_CLIENT (self));
	g_signal_emit_by_name (G_OBJECT (self), "connected", socket_fd);
}


void snul_tcp_client_on_data_received (SnulTcpClient* self, const char* buffer, gsize len) {
	g_return_if_fail (SNUL_IS_TCP_CLIENT (self));
	g_signal_emit_by_name (G_OBJECT (self), "data-received", buffer, len);
}


void snul_tcp_client_on_disconnected (SnulTcpClient* self) {
	g_return_if_fail (SNUL_IS_TCP_CLIENT (self));
	g_signal_emit_by_name (G_OBJECT (self), "disconnected");
}


void snul_tcp_client_on_error (SnulTcpClient* self, GError* error) {
	g_return_if_fail (SNUL_IS_TCP_CLIENT (self));
	g_signal_emit_by_name (G_OBJECT (self), "error", error);
}


static void snul_tcp_client_on_connected_real (SnulTcpClient* self, GError** error) {
	g_return_if_fail (SNUL_IS_TCP_CLIENT (self));
}


static gboolean snul_tcp_client_socket_callback (GIOChannel* channel, GIOCondition condition, gpointer data) {
	GError * inner_error;
	SnulTcpClient* _tmp0;
	SnulTcpClient* client;
	gboolean res;
	GIOCondition _tmp1;
	gboolean _tmp2;
	inner_error = NULL;
	_tmp0 = NULL;
	client = (_tmp0 = SNUL_TCP_CLIENT (data), (_tmp0 == NULL ? NULL : g_object_ref (_tmp0)));
	res = TRUE;
	_tmp1 = condition;
	if (_tmp1 == G_IO_IN)
	do {
		gsize bytes_read;
		bytes_read = ((gsize) 0);
		{
			g_io_channel_read_chars (channel, snul_tcp_client__io_channel_buffer, ((gsize) snul_tcp_client_BUFFER_LENGTH - 1), &bytes_read, &inner_error);
			if (inner_error != NULL) {
				goto __catch7_g_error;
			}
			/*null terminate*/
			snul_tcp_client__io_channel_buffer[bytes_read] = ((gchar) NULL);
			if (bytes_read > 0) {
				snul_tcp_client_on_data_received (client, ((char*) snul_tcp_client__io_channel_buffer), bytes_read);
			} else {
				snul_tcp_client_on_error (client, g_error_new (SNUL_TCP_SOCKET_ERROR, SNUL_TCP_SOCKET_ERROR_SOCKET_BROKEN, "zero length data, socket broken? Closing connection."));
				snul_tcp_client_disconnect (client);
			}
		}
		goto __finally7;
		__catch7_g_error:
		{
			GError * err;
			err = inner_error;
			inner_error = NULL;
			{
				g_warning ("snul.vala:262: error reading: %s", err->message);
			}
		}
		__finally7:
		;
		break;
	} while (0); else if (_tmp1 == G_IO_HUP)
	do {
		snul_tcp_client_on_error (client, g_error_new (SNUL_TCP_SOCKET_ERROR, SNUL_TCP_SOCKET_ERROR_SOCKET_HUP, "Socket HUP"));
		snul_tcp_client_disconnect (client);
		res = FALSE;
		break;
	} while (0); else if (_tmp1 == G_IO_ERR)
	do {
		snul_tcp_client_on_error (client, g_error_new (SNUL_TCP_SOCKET_ERROR, SNUL_TCP_SOCKET_ERROR_IOCHANNEL_ERROR, "IOChannel error (ERR)"));
		snul_tcp_client_disconnect (client);
		break;
	} while (0); else if (_tmp1 == G_IO_NVAL)
	do {
		snul_tcp_client_on_error (client, g_error_new (SNUL_TCP_SOCKET_ERROR, SNUL_TCP_SOCKET_ERROR_IOCHANNEL_NVAL, "IOChannel invalid request (NAVL)"));
		snul_tcp_client_disconnect (client);
		break;
	} while (0);
	return (_tmp2 = res, (client == NULL ? NULL : (client = (g_object_unref (client), NULL))), _tmp2);
}


SnulTcpClient* snul_tcp_client_new (void) {
	SnulTcpClient * self;
	self = g_object_newv (SNUL_TYPE_TCP_CLIENT, 0, NULL);
	return self;
}


char* snul_tcp_client_get_address (SnulTcpClient* self) {
	g_return_val_if_fail (SNUL_IS_TCP_CLIENT (self), NULL);
	return self->priv->_address;
}


char* snul_tcp_client_get_first_address (SnulTcpClient* self) {
	g_return_val_if_fail (SNUL_IS_TCP_CLIENT (self), NULL);
	return snul_tcp_client_get_address (self);
}


void snul_tcp_client_set_first_address (SnulTcpClient* self, const char* value) {
	char* _tmp2;
	const char* _tmp1;
	g_return_if_fail (SNUL_IS_TCP_CLIENT (self));
	_tmp2 = NULL;
	_tmp1 = NULL;
	self->priv->_first_address = (_tmp2 = (_tmp1 = value, (_tmp1 == NULL ? NULL : g_strdup (_tmp1))), (self->priv->_first_address = (g_free (self->priv->_first_address), NULL)), _tmp2);
}


char* snul_tcp_client_get_port (SnulTcpClient* self) {
	g_return_val_if_fail (SNUL_IS_TCP_CLIENT (self), NULL);
	return self->priv->_port;
}


static void snul_tcp_client_get_property (GObject * object, guint property_id, GValue * value, GParamSpec * pspec) {
	SnulTcpClient * self;
	self = SNUL_TCP_CLIENT (object);
	switch (property_id) {
		case SNUL_TCP_CLIENT_ADDRESS:
		g_value_set_string (value, snul_tcp_client_get_address (self));
		break;
		case SNUL_TCP_CLIENT_FIRST_ADDRESS:
		g_value_set_string (value, snul_tcp_client_get_first_address (self));
		break;
		case SNUL_TCP_CLIENT_PORT:
		g_value_set_string (value, snul_tcp_client_get_port (self));
		break;
		default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}


static void snul_tcp_client_set_property (GObject * object, guint property_id, const GValue * value, GParamSpec * pspec) {
	SnulTcpClient * self;
	self = SNUL_TCP_CLIENT (object);
	switch (property_id) {
		case SNUL_TCP_CLIENT_FIRST_ADDRESS:
		snul_tcp_client_set_first_address (self, g_value_get_string (value));
		break;
		default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
		break;
	}
}


static void snul_tcp_client_class_init (SnulTcpClientClass * klass) {
	snul_tcp_client_parent_class = g_type_class_peek_parent (klass);
	g_type_class_add_private (klass, sizeof (SnulTcpClientPrivate));
	G_OBJECT_CLASS (klass)->get_property = snul_tcp_client_get_property;
	G_OBJECT_CLASS (klass)->set_property = snul_tcp_client_set_property;
	G_OBJECT_CLASS (klass)->dispose = snul_tcp_client_dispose;
	g_object_class_install_property (G_OBJECT_CLASS (klass), SNUL_TCP_CLIENT_ADDRESS, g_param_spec_string ("address", "address", "address", NULL, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), SNUL_TCP_CLIENT_FIRST_ADDRESS, g_param_spec_string ("first-address", "first-address", "first-address", NULL, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE | G_PARAM_WRITABLE));
	g_object_class_install_property (G_OBJECT_CLASS (klass), SNUL_TCP_CLIENT_PORT, g_param_spec_string ("port", "port", "port", NULL, G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK | G_PARAM_STATIC_BLURB | G_PARAM_READABLE));
	g_signal_new ("connected", SNUL_TYPE_TCP_CLIENT, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_marshal_VOID__INT, G_TYPE_NONE, 1, G_TYPE_INT);
	g_signal_new ("data_received", SNUL_TYPE_TCP_CLIENT, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_user_marshal_VOID__STRING_ULONG, G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_ULONG);
	g_signal_new ("disconnected", SNUL_TYPE_TCP_CLIENT, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_marshal_VOID__VOID, G_TYPE_NONE, 0);
	g_signal_new ("error", SNUL_TYPE_TCP_CLIENT, G_SIGNAL_RUN_LAST, 0, NULL, NULL, g_cclosure_marshal_VOID__POINTER, G_TYPE_NONE, 1, G_TYPE_POINTER);
	snul_tcp_client__io_channel_buffer = g_new0 (gchar, snul_tcp_client_BUFFER_LENGTH);
}


static void snul_tcp_client_init (SnulTcpClient * self) {
	self->priv = SNUL_TCP_CLIENT_GET_PRIVATE (self);
	self->priv->_address = NULL;
	self->priv->_first_address = NULL;
	self->priv->_port = NULL;
	self->priv->_port_number = ((gushort) 0);
	self->priv->_socket_fd = 0;
	self->priv->_io_channel = NULL;
	self->priv->_watch_id = ((guint) 0);
	self->priv->_resolver_watch_id = ((guint) 0);
	self->priv->_resolving_address = FALSE;
	self->priv->_resolv_timeout = ((gboolean) 30);
	self->priv->hp = NULL;
}


static void snul_tcp_client_dispose (GObject * obj) {
	SnulTcpClient * self;
	self = SNUL_TCP_CLIENT (obj);
	self->priv->_address = (g_free (self->priv->_address), NULL);
	self->priv->_first_address = (g_free (self->priv->_first_address), NULL);
	self->priv->_port = (g_free (self->priv->_port), NULL);
	(self->priv->_io_channel == NULL ? NULL : (self->priv->_io_channel = (g_io_channel_unref (self->priv->_io_channel), NULL)));
	G_OBJECT_CLASS (snul_tcp_client_parent_class)->dispose (obj);
}


GType snul_tcp_client_get_type (void) {
	static GType snul_tcp_client_type_id = 0;
	if (G_UNLIKELY (snul_tcp_client_type_id == 0)) {
		static const GTypeInfo g_define_type_info = { sizeof (SnulTcpClientClass), (GBaseInitFunc) NULL, (GBaseFinalizeFunc) NULL, (GClassInitFunc) snul_tcp_client_class_init, (GClassFinalizeFunc) NULL, NULL, sizeof (SnulTcpClient), 0, (GInstanceInitFunc) snul_tcp_client_init };
		snul_tcp_client_type_id = g_type_register_static (G_TYPE_OBJECT, "SnulTcpClient", &g_define_type_info, 0);
	}
	return snul_tcp_client_type_id;
}



static void g_cclosure_user_marshal_VOID__STRING_ULONG (GClosure * closure, GValue * return_value, guint n_param_values, const GValue * param_values, gpointer invocation_hint, gpointer marshal_data) {
	typedef void (*GMarshalFunc_VOID__STRING_ULONG) (gpointer data1, const char* arg_1, gsize arg_2, gpointer data2);
	register GMarshalFunc_VOID__STRING_ULONG callback;
	register GCClosure * cc;
	register gpointer data1, data2;
	cc = ((GCClosure *) closure);
	g_return_if_fail (n_param_values == 3);
	if (G_CCLOSURE_SWAP_DATA (closure)) {
		data1 = closure->data;
		data2 = param_values->data[0].v_pointer;
	} else {
		data1 = param_values->data[0].v_pointer;
		data2 = closure->data;
	}
	callback = ((GMarshalFunc_VOID__STRING_ULONG) (marshal_data ? marshal_data : cc->callback));
	callback (data1, g_value_get_string (param_values + 1), g_value_get_ulong (param_values + 2), data2);
}



